#!/usr/local/bin/ruby
# vi:set ts=3 sw=3:
# vim:set sts=0 noet:

=begin
Copyright (c) 2009 KISHIMOTO, Makoto

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
=end

require 'pp'

require 'tempfile'
require 'webrick'
require 'image_size'

require 'rubygems'

require 'extexif'

module ViewJPEG
  class ExFileHandler < WEBrick::HTTPServlet::FileHandler
    include WEBrick
    # from class FileHandler, in webrick/httpservlet/filehandler.rb
    def set_dir_list(req, res)
        redirect_to_directory_uri(req, res)
        unless @options[:FancyIndexing]
          raise HTTPStatus::Forbidden, "no access permission to `#{req.path}'"
        end
        local_path = res.filename
        list = Dir::entries(local_path).collect{|name|
          next if name == "." || name == ".."
          next if nondisclosure_name?(name)
          next if windows_ambiguous_name?(name)
          st = (File::stat(File.join(local_path, name)) rescue nil)
          if st.nil?
            [ name, nil, -1 ]
          elsif st.directory?
            [ name + "/", st.mtime, -1 ]
          else
            [ name, st.mtime, st.size ]
          end
        }
        list.compact!

        if    d0 = req.query["N"]; idx = 0
        elsif d0 = req.query["M"]; idx = 1
        elsif d0 = req.query["S"]; idx = 2
        else  d0 = "A"           ; idx = 0
        end
        d1 = (d0 == "A") ? "D" : "A"

        if d0 == "A"
          list.sort!{|a,b| a[idx] <=> b[idx] }
        else
          list.sort!{|a,b| b[idx] <=> a[idx] }
        end

        res['content-type'] = "text/html"

        res.body = <<-_end_of_html_
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 3.2 Final//EN">
<HTML>
  <HEAD><TITLE>Index of #{HTMLUtils::escape(req.path)}</TITLE></HEAD>
  <BODY>
    <H1>Index of #{HTMLUtils::escape(req.path)}</H1>
        _end_of_html_

        res.body << "<PRE>\n"
        res.body << " <A HREF=\"?N=#{d1}\">Name</A>                          "
        res.body << "<A HREF=\"?M=#{d1}\">Last modified</A>         "
        res.body << "<A HREF=\"?S=#{d1}\">Size</A>\n"
        res.body << "<HR>\n"
       
        list.unshift [ "..", File::mtime(local_path+"/.."), -1 ]
        list.each{ |name, time, size|
          if name == ".."
            dname = "Parent Directory"
          elsif name.size > 25
            dname = name.sub(/^(.{23})(.*)/){ $1 + ".." }
          else
            dname = name
          end
          # modified for ExFileHandler from here
          s =  " <A HREF=\"#{HTTPUtils::escape(name)}\">"
          if /\.jpe?g\z/i.match name then
            make_jpegentry s, name
          end
          s <<  "#{HTMLUtils::escape(dname)}</A>"
          # modified for ExFileHandler to here
          s << " " * (30 - dname.size)
          s << (time ? time.strftime("%Y/%m/%d %H:%M      ") : " " * 22)
          s << (size >= 0 ? size.to_s : "-") << "\n"
          res.body << s
        }
        res.body << "</PRE><HR>"

        res.body << <<-_end_of_html_    
    <ADDRESS>
     #{HTMLUtils::escape(@config[:ServerSoftware])}<BR>
     at #{req.host}:#{req.port}
    </ADDRESS>
  </BODY>
</HTML>
        _end_of_html_
     end

     def make_jpegentry s, name
       s << "<IMG SRC=\"#{HTTPUtils::escape(name)}?thumb\">\n "
     end
  end

	class JpegHandler < WEBrick::HTTPServlet::DefaultFileHandler
		TRANSFORM_TABLE = {
			'top - left' => nil,
			'top - right' => '-lr',
			'bottom - right' => '-r180',
			'bottom - left' => '-tb',
			'left - top' => '-xy',
			'right - top' => '-cw',
			'right - bottom' => '-xform=leftright,topbottom,transpose',
			'left - bottom' => '-ccw'}

		def do_GET req, res
			exif = ExtExif.new @local_path

			orientation = exif['Orientation']
			transform = TRANSFORM_TABLE[orientation]

			begin
				thumb_data = ''
				exif.thumbnail thumb_data
			rescue Exif::Error::ThumbnailNotFound
				thumb_data = nil
			end

			st = File::stat @local_path
			mtime = st.mtime
			res['etag'] = sprintf "%x-%x-%x", st.ino, st.size, st.mtime.to_i

			if not_modified? req, res, mtime, res['etag'] then
				res.body = ''
				raise WEBrick::HTTPStatus::NotModified
			end

			thumb = (req.query_string == 'thumb')
			if transform and thumb then
				mtype = WEBrick::HTTPUtils::mime_type @local_path, @config[:MimeTypes]
				res['content-type'] = mtype
				res['last-modified'] = mtime.httpdate
				if thumb_data then
					result = Tempfile.open 'viewjpeg'
					result.close
					opipes = IO.pipe
					unless pid = fork then
						opipes[1].close
						STDIN.close
						opipes[0].dup
						opipes[0].close
						exec '/bin/sh', '-c', "jpegtopnm | pamflip #{transform} | pnmtojpeg > '#{result.path}'"
					end
					opipes[0].close
					opipes[1].write thumb_data
					opipes[1].flush
					opipes[1].close
					Process.waitpid pid
				else
					image_size = nil
					open @local_path do|fp|
						image_size = ImageSize.new fp
					end
					scaleoption = if image_size.get_width > image_size.get_height then '-xsize=160' else '-ysize=160' end

					result = Tempfile.open 'viewjpeg'
					result.close
					system '/bin/sh', '-c', "jpegtopnm '#{@local_path}' | pamscale #{scaleoption} | pamflip #{transform} | pnmtojpeg > '#{result.path}'"
				end
				st = File::stat result.path
				res['content-length'] = st.size
				res.body = result.open
				result.delete
			elsif transform then
				mtype = WEBrick::HTTPUtils::mime_type @local_path, @config[:MimeTypes]
				res['content-type'] = mtype
				res['last-modified'] = mtime.httpdate

				result = Tempfile.open 'viewjpeg'
				result.close
				system '/bin/sh', '-c', "jpegtopnm '#{@local_path}' | pamflip #{transform} | pnmtojpeg > '#{result.path}'"

				st = File::stat result.path
				res['content-length'] = st.size
				res.body = result.open
				result.delete
			elsif thumb then
				mtype = WEBrick::HTTPUtils::mime_type @local_path, @config[:MimeTypes]
				res['content-type'] = mtype
				res['last-modified'] = mtime.httpdate
				if thumb_data then
					res['content-length'] = thumb_data.size
					res.body = thumb_data
				else
					image_size = nil
					open @local_path do|fp|
						image_size = ImageSize.new fp
					end
					scaleoption = if image_size.get_width > image_size.get_height then '-xsize=160' else '-ysize=160' end

					result = Tempfile.open 'viewjpeg'
					result.close
					system '/bin/sh', '-c', "jpegtopnm '#{@local_path}' | pamscale #{scaleoption} | pnmtojpeg > '#{result.path}'"

					st = File::stat result.path
					res['content-length'] = st.size
					res.body = result.open
					result.delete
				end
			else
				super
			end
		end
	end
end

handler_table = {'jpg' => ViewJPEG::JpegHandler, 'jpeg' => ViewJPEG::JpegHandler }
document_root_options = { :FancyIndexing => true, :HandlerTable => handler_table }
server = WEBrick::HTTPServer.new :Port => 8080
server.mount '/', ViewJPEG::ExFileHandler, 'MyPhoto', document_root_options
trap 'INT' do server.shutdown end
server.start
